@@ -14,7 +14,6 @@
 """Object Serialization for Mongo/BSON"""
 from __future__ import absolute_import
 import copy_reg
-import struct
 
 import lru
 import persistent.interfaces
@@ -29,16 +28,14 @@ from zope.dottedname.resolve import resolve
 
 from mongopersist import interfaces
 
+IGNORE_IDENTICAL_DOCUMENTS = True
+ALWAYS_READ_FULL_DOC = True
+
 SERIALIZERS = []
 OID_CLASS_LRU = lru.LRUCache(20000)
-
-def p64(v):
-    """Pack an integer or long into a 8-byte string"""
-    return struct.pack(">Q", v)
-
-def u64(v):
-    """Unpack an 8-byte string into a 64-bit long integer."""
-    return struct.unpack(">Q", v)[0]
+COLLECTIONS_WITH_TYPE = set()
+AVAILABLE_NAME_MAPPINGS = set()
+PATH_RESOLVE_CACHE = {}
 
 def get_dotted_name(obj):
     return obj.__module__+'.'+obj.__name__
@@ -80,11 +77,15 @@ class ObjectWriter(object):
         except AttributeError:
             return db_name, get_dotted_name(obj.__class__)
         # Make sure that the coll_name to class path mapping is available.
+        # Let's make sure we do the lookup only once, since the info will
+        # never change.
+        path = get_dotted_name(obj.__class__)
+        map = {'collection': coll_name, 'database': db_name, 'path': path}
+        map_hash = (db_name, coll_name, path)
+        if map_hash in AVAILABLE_NAME_MAPPINGS:
+            return db_name, coll_name
         db = self._jar._conn[self._jar.default_database]
         coll = db[self._jar.name_map_collection]
-        map = {'collection': coll_name,
-               'database': db_name,
-               'path': get_dotted_name(obj.__class__)}
         result = coll.find_one(map)
         if result is None:
             # If there is already a map for this collection, the next map must
@@ -95,6 +96,7 @@ class ObjectWriter(object):
                 setattr(obj, '_p_mongo_store_type', True)
             map['doc_has_type'] = getattr(obj, '_p_mongo_store_type', False)
             coll.save(map)
+        AVAILABLE_NAME_MAPPINGS.add(map_hash)
         return db_name, coll_name
 
     def get_non_persistent_state(self, obj, seen):
@@ -211,6 +213,20 @@ class ObjectWriter(object):
 
         return self.get_non_persistent_state(obj, seen)
 
+    def get_full_state(self, obj):
+        doc = self.get_state(obj.__getstate__())
+        # Add a persistent type info, if necessary.
+        if getattr(obj, '_p_mongo_store_type', False):
+            doc['_py_persistent_type'] = get_dotted_name(obj.__class__)
+        # A hook, so that the conflict handler can modify the state document
+        # if needed.
+        self._jar.conflict_handler.on_before_store(obj, doc)
+        # Add the object id.
+        if obj._p_oid is not None:
+            doc['_id'] = obj._p_oid.id
+        # Return the full state document
+        return doc
+
     def store(self, obj, ref_only=False):
         __traceback_info__ = (obj, ref_only)
 
@@ -229,14 +245,15 @@ class ObjectWriter(object):
             doc = self.get_state(obj.__getstate__())
         if getattr(obj, '_p_mongo_store_type', False):
             doc['_py_persistent_type'] = get_dotted_name(obj.__class__)
-        # If conflict detection is turned on, store a serial number for the
-        # document.
-        if self._jar.detect_conflicts:
-            doc['_py_serial'] = u64(getattr(obj, '_p_serial', 0)) + 1
-            obj._p_serial = p64(doc['_py_serial'])
 
+        # A hook, so that the conflict handler can modify the state document
+        # if needed.
+        self._jar.conflict_handler.on_before_store(obj, doc)
+
+        stored = False
         if obj._p_oid is None:
             doc_id = coll.insert(doc)
+            stored = True
             obj._p_jar = self._jar
             obj._p_oid = pymongo.dbref.DBRef(coll_name, doc_id, db_name)
             # Make sure that any other code accessing this object in this
@@ -244,7 +261,24 @@ class ObjectWriter(object):
             self._jar._object_cache[doc_id] = obj
         else:
             doc['_id'] = obj._p_oid.id
-            coll.save(doc)
+            # We only want to store a new version of the document, if it is
+            # different. We have to delegate that task to the conflict
+            # handler, since it might know about meta-fields that need to be
+            # ignored.
+            orig_doc = self._jar._latest_states.get(obj._p_oid)
+            if (not IGNORE_IDENTICAL_DOCUMENTS or
+                not self._jar.conflict_handler.is_same(obj, orig_doc, doc)):
+                coll.save(doc)
+                stored = True
+
+        if stored:
+            # Make sure that the doc is added to the latest states.
+            self._jar._latest_states[obj._p_oid] = doc
+
+            # A hook, so that the conflict handler can modify the object or state
+            # document after an object was stored.
+            self._jar.conflict_handler.on_after_store(obj, doc)
+
         return obj._p_oid
 
 
@@ -253,45 +287,116 @@ class ObjectReader(object):
 
     def __init__(self, jar):
         self._jar = jar
+        self._single_map_cache = {}
 
     def simple_resolve(self, path):
-        return resolve(path)
+        # We try to look up the klass from a cache. The important part here is
+        # that we also cache lookup failures as None, since they actually
+        # happen more frequently than a hit due to an optimization in the
+        # resolve() function.
+        try:
+            klass = PATH_RESOLVE_CACHE[path]
+        except KeyError:
+            try:
+                klass = resolve(path)
+            except ImportError:
+                PATH_RESOLVE_CACHE[path] = klass = None
+            else:
+                PATH_RESOLVE_CACHE[path] = klass
+        if klass is None:
+            raise ImportError(path)
+        return klass
 
     def resolve(self, dbref):
         __traceback_info__ = dbref
+        # 1. Check the global oid-based lookup cache. Use the hash of the id,
+        #    since otherwise the comparison is way too expensive.
         try:
-            return OID_CLASS_LRU[dbref.id]
+            return OID_CLASS_LRU[hash(dbref.id)]
         except KeyError:
             pass
-        # First we try to resolve the path directly.
+        # 2. Check the transient single map entry lookup cache.
+        try:
+            return self._single_map_cache[(dbref.database, dbref.collection)]
+        except KeyError:
+            pass
+        # 3. If we have found the type within the document for a collection
+        #    before, let's try again. This will only hit, if we have more than
+        #    one type for the collection, otherwise the single map entry
+        #    lookup failed.
+        coll_key = (dbref.database, dbref.collection)
+        if coll_key in COLLECTIONS_WITH_TYPE:
+            if dbref in self._jar._latest_states:
+                obj_doc = self._jar._latest_states[dbref]
+            elif ALWAYS_READ_FULL_DOC:
+                obj_doc = self._jar.get_collection(
+                    dbref.database, dbref.collection).find_one(dbref.id)
+                self._jar._latest_states[dbref] = obj_doc
+            else:
+                obj_doc = self._jar\
+                    .get_collection(dbref.database, dbref.collection)\
+                    .find_one(dbref.id, fields=('_py_persistent_type',))
+            if '_py_persistent_type' in obj_doc:
+                klass = self.simple_resolve(obj_doc['_py_persistent_type'])
+                OID_CLASS_LRU[hash(dbref.id)] = klass
+                return klass
+        # 4. Try to resolve the path directly. We want to do this optimization
+        #    after all others, because trying it a lot is very expensive.
         try:
             return self.simple_resolve(dbref.collection)
         except ImportError:
             pass
+        # 5. No simple hits, so we have to do some leg work.
         # Let's now try to look up the path from the collection to path
         # mapping
         db = self._jar._conn[self._jar.default_database]
         coll = db[self._jar.name_map_collection]
-        result = coll.find(
-            {'collection': dbref.collection, 'database': dbref.database})
-        if result.count() == 0:
+        result = tuple(coll.find(
+            {'collection': dbref.collection, 'database': dbref.database}))
+        # Calling count() on a query result causes another database
+        # access. Since the result sets should be typically very small, let's
+        # load them all.
+        count = len(result)
+        if count == 0:
             raise ImportError(dbref)
-        elif result.count() == 1:
+        elif count == 1:
             # Do not add these results to the LRU cache, since the count might
-            # change later.
-            return self.simple_resolve(result.next()['path'])
+            # change later. But storing it for the length of the transaction
+            # is fine, which is really useful if you load a lot of objects of
+            # the same type.
+            klass = self.simple_resolve(result[0]['path'])
+            self._single_map_cache[(dbref.database, dbref.collection)] = klass
+            return klass
         else:
             if dbref.id is None:
                 raise ImportError(dbref)
             # Multiple object types are stored in the collection. We have to
             # look at the object to find out the type.
-            obj_doc = self._jar\
-                .get_collection(dbref.database, dbref.collection).find_one(
-                    dbref.id, fields=('_py_persistent_type',))
+            if dbref in self._jar._latest_states:
+                # Optimization: If we have the latest state, then we just get
+                # this object document. This is used for fast loading or when
+                # resolving the same object path a second time. (The latter
+                # should never happen due to the object cache.)
+                obj_doc = self._jar._latest_states[dbref]
+            elif ALWAYS_READ_FULL_DOC:
+                # Optimization: Read the entire doc and stick it in the right
+                # place so that unghostifying the object later will not cause
+                # another database access.
+                obj_doc = self._jar\
+                    .get_collection(dbref.database, dbref.collection)\
+                    .find_one(dbref.id)
+                self._jar._latest_states[dbref] = obj_doc
+            else:
+                obj_doc = self._jar\
+                    .get_collection(dbref.database, dbref.collection)\
+                    .find_one(dbref.id, fields=('_py_persistent_type',))
             if '_py_persistent_type' in obj_doc:
+                COLLECTIONS_WITH_TYPE.add(coll_key)
                 klass = self.simple_resolve(obj_doc['_py_persistent_type'])
             else:
                 # Find the name-map entry where "doc_has_type" is False.
+                # Note: This case is really inefficient and does not allow any
+                # optimization. It should be avoided as much as possible.
                 for name_map_item in result:
                     if not name_map_item['doc_has_type']:
                         klass = self.simple_resolve(name_map_item['path'])
@@ -384,17 +489,27 @@ class ObjectReader(object):
             coll = self._jar.get_collection(
                 obj._p_oid.database, obj._p_oid.collection)
             doc = coll.find_one({'_id': obj._p_oid.id})
+        # Create a copy of the doc, so that we can modify it.
+        state_doc = doc.copy()
         # Remove unwanted attributes.
-        doc.pop('_id')
-        doc.pop('_py_persistent_type', None)
-        # Store the serial, if conflict detection is enabled.
-        if self._jar.detect_conflicts:
-            obj._p_serial = p64(doc.pop('_py_serial', 0))
+        state_doc.pop('_id')
+        state_doc.pop('_py_persistent_type', None)
+        # Allow the conflict handler to modify the object or state document
+        # before it is set on the object.
+        self._jar.conflict_handler.on_before_set_state(obj, state_doc)
         # Now convert the document to a proper Python state dict.
-        state = dict(self.get_object(doc, obj))
+        state = dict(self.get_object(state_doc, obj))
         # Now store the original state. It is assumed that the state dict is
         # not modified later.
-        self._jar._original_states[obj._p_oid] = doc
+        # Make sure that we never set the original state multiple times, even
+        # if reassigning the state within the same transaction. Otherwise we
+        # can never fully undo a transaction.
+        if obj._p_oid not in self._jar._original_states:
+            self._jar._original_states[obj._p_oid] = doc
+            # Sometimes this method is called to update the object state
+            # before storage. Only update the latest states when the object is
+            # originally loaded.
+            self._jar._latest_states[obj._p_oid] = doc
         # Set the state.
         obj.__setstate__(state)
 
