@@ -14,11 +14,10 @@
 """Mongo  Tests"""
 import doctest
 import persistent
-import pprint
 import transaction
 from pymongo import dbref, objectid
 
-from mongopersist import interfaces, testing, datamanager
+from mongopersist import conflict, interfaces, testing, datamanager
 
 class Foo(persistent.Persistent):
     def __init__(self, name=None):
@@ -27,19 +26,6 @@ class Foo(persistent.Persistent):
 class Bar(persistent.Persistent):
     _p_mongo_sub_object = True
 
-def doctest_create_conflict_error():
-    r"""create_conflict_error(): General Test
-
-    Simple helper function to create a conflict error.
-
-     >>> foo = Foo()
-     >>> foo._p_serial = '\x00\x00\x00\x00\x00\x00\x00\x01'
-
-     >>> datamanager.create_conflict_error(foo, {'_py_serial': 3})
-     ConflictError: database conflict error
-                    (oid None, class Foo, start serial 1, current serial 3)
-    """
-
 def doctest_Root():
     r"""Root: General Test
 
@@ -172,12 +158,10 @@ def doctest_MongoDataManager_object_dump_load_reset():
 
       >>> dm = datamanager.MongoDataManager(
       ...     conn,
-      ...     detect_conflicts=True,
       ...     default_database = DBNAME,
       ...     root_database = DBNAME,
       ...     root_collection = 'proot',
-      ...     name_map_collection = 'coll_pypath_map',
-      ...     conflict_error_factory = datamanager.create_conflict_error)
+      ...     name_map_collection = 'coll_pypath_map')
 
     There are two convenience methods that let you serialize and de-serialize
     objects explicitly:
@@ -217,6 +201,70 @@ def doctest_MongoDataManager_object_dump_load_reset():
       >>> foo._p_oid = foo2._p_oid
     """
 
+def doctest_MongoDataManager_dump_only_on_real_change():
+    r"""MongoDataManager: dump(): dump on real change only.
+
+    The data manager only writes data when we actually have a difference in
+    state.
+
+    We have to use a serial conflict handler, otherwise it is hard to check
+    whether data was written.
+
+      >>> dm.conflict_handler = conflict.SimpleSerialConflictHandler(dm)
+
+    Let's now add an object:
+
+      >>> foo = Foo('foo')
+      >>> foo_ref = dm.insert(foo)
+      >>> dm.tpc_finish(None)
+
+      >>> coll = dm._get_collection_from_object(foo)
+      >>> coll.find_one({})
+      {u'_id': ObjectId('...'), u'_py_serial': 1, u'name': u'foo'}
+
+    So the original state is in. Let's now modify an object:
+
+      >>> foo = dm.load(foo_ref)
+      >>> foo.name = 'Foo'
+      >>> foo._p_changed
+      True
+      >>> dm.tpc_finish(None)
+
+      >>> coll.find_one({})
+      {u'_id': ObjectId('...'), u'_py_serial': 2, u'name': u'Foo'}
+
+    If we now modify the object again, but write the same value, the state
+    should not be written to Mongo.
+
+      >>> foo = dm.load(foo_ref)
+      >>> foo.name = 'Foo'
+      >>> foo._p_changed
+      True
+      >>> dm.tpc_finish(None)
+
+      >>> coll.find_one({})
+      {u'_id': ObjectId('...'), u'_py_serial': 2, u'name': u'Foo'}
+
+    Let's make sure everything also works when we flush the transaction in the
+    middle.
+
+      >>> foo = dm.load(foo_ref)
+      >>> foo.name = 'fuh'
+      >>> dm.flush()
+      >>> coll.find_one({})
+      {u'_id': ObjectId('...'), u'_py_serial': 3, u'name': u'fuh'}
+
+      >>> foo._p_changed
+      False
+      >>> foo.name = 'fuh'
+      >>> foo._p_changed
+      True
+
+      >>> dm.tpc_finish(None)
+      >>> coll.find_one({})
+      {u'_id': ObjectId('...'), u'_py_serial': 3, u'name': u'fuh'}
+    """
+
 def doctest_MongoDataManager_flush():
     r"""MongoDataManager: flush()
 
@@ -226,7 +274,7 @@ def doctest_MongoDataManager_flush():
 
     We also want to test the effects of conflict detection:
 
-      >>> dm.detect_conflicts = True
+      >>> dm.conflict_handler = conflict.SimpleSerialConflictHandler(dm)
 
     Let's now add an object to the database and reset the manager like it is
     done at the end of a transaction:
@@ -283,9 +331,10 @@ def doctest_MongoDataManager_flush():
         ...
         ConflictError: database conflict error
             (oid DBRef('mongopersist.tests.test_datamanager.Foo',
-                       ObjectId('4f5bfcaf37a08e2849000000'),
+                       ObjectId('4e7ddf12e138237403000000'),
                        'mongopersist_test'),
-             class Foo, start serial 1, current serial 2)
+             class Foo,
+             orig serial 1, cur serial 2, new serial 2)
     """
 
 def doctest_MongoDataManager_insert():
@@ -339,6 +388,25 @@ def doctest_MongoDataManager_insert():
        {u'_id': ObjectId('4f5c443837a08e37bf000001'), u'name': u'Foo 2'})
     """
 
+def doctest_MongoDataManager_insert_conflict_detection():
+    r"""MongoDataManager: insert(obj): Conflict Detection.
+
+    This test ensures that if the datamanager has conflict detection turned
+    on, all the needed helper fields are written.
+
+      >>> dm.conflict_handler = conflict.SimpleSerialConflictHandler(dm)
+      >>> foo = Foo('foo')
+      >>> foo_ref = dm.insert(foo)
+
+    Let's check that all the fields are there:
+
+      >>> coll = dm.get_collection_from_object(foo)
+      >>> coll.find_one({})
+      {u'_id': ObjectId('4f74837237a08e186f000000'), u'_py_serial': 1,
+       u'name': u'foo'}
+    """
+
+
 def doctest_MongoDataManager_remove():
     r"""MongoDataManager: remove(obj)
 
@@ -508,72 +576,126 @@ def doctest_MongoDataManager_abort():
        {u'_id': ObjectId('4f5c114f37a08e2cac000001'), u'name': u'two'})
     """
 
-def doctest_MongoDataManager_commit():
-    r"""MongoDataManager: commit()
+def doctest_MongoDataManager_abort_modified_only():
+    r"""MongoDataManager: abort(): Only reset changed objects.
 
-    Contrary to what the name suggests, this is the commit called during the
-    first phase of a two-phase commit. Thus, for all practically purposes,
-    this method merely checks whether the commit would potentially fail.
+    We want to make sure that we only reset modified objects, not all objects
+    that have been loaded. The ratio from reads to writes is very high, so
+    unexpected behavior with other transactions is decreased by that ratio.
 
-    This means, if conflict detection is disabled, this method does nothing.
-
-      >>> dm.detect_conflicts
-      False
-      >>> dm.commit(transaction.get())
-
-    Let's now turn on conflict detection:
-
-      >>> dm.detect_conflicts = True
-
-    For new objects (not having an oid), it always passes:
+    First let's create an initial state:
 
       >>> dm.reset()
-      >>> dm._registered_objects = [Foo()]
-      >>> dm.commit(transaction.get())
-
-    If the object has an oid, but is not found in the DB, we also just pass,
-    because the object will be inserted.
-
-      >>> foo = Foo()
-      >>> foo._p_oid =  dbref.DBRef(
-      ...     'mongopersist.tests.test_datamanager.Foo',
-      ...     objectid.ObjectId('4eb2eb7437a08e0156000000'),
-      ...     'mongopersist_test')
-
+      >>> foo1_ref = dm.insert(Foo('one'))
+      >>> foo2_ref = dm.insert(Foo('two'))
+      >>> foo3_ref = dm.insert(Foo('three'))
       >>> dm.reset()
-      >>> dm._registered_objects = [foo]
-      >>> dm.commit(transaction.get())
 
-    Let's now store an object and make sure it does not conflict:
+      >>> coll = dm._get_collection_from_object(Foo())
+      >>> tuple(coll.find({}))
+      ({u'_id': ObjectId('4f5c114f37a08e2cac000000'), u'name': u'one'},
+       {u'_id': ObjectId('4f5c114f37a08e2cac000001'), u'name': u'two'},
+       {u'_id': ObjectId('4f5c114f37a08e2cac000002'), u'name': u'three'})
 
-      >>> foo = Foo()
-      >>> ref = dm.dump(foo)
-      >>> ref
-      DBRef('mongopersist.tests.test_datamanager.Foo',
-            ObjectId('4eb3468037a08e1b74000000'),
-            'mongopersist_test')
+    1. Transaction A loads all objects:
 
+        >>> foo1_A = dm.load(foo1_ref)
+        >>> foo1_A.name
+        u'one'
+        >>> foo2_A = dm.load(foo2_ref)
+        >>> foo2_A.name
+        u'two'
+        >>> foo3_A = dm.load(foo3_ref)
+        >>> foo3_A.name
+        u'three'
+
+        >>> sorted([ref.id for ref in dm._original_states.keys()])
+        [ObjectId('4f746d0b37a08e1013000000'),
+         ObjectId('4f746d0b37a08e1013000001'),
+         ObjectId('4f746d0b37a08e1013000002')]
+
+    2. Transaction B comes along and modifies Foo 3's data and commits:
+
+        >>> dm_B = datamanager.MongoDataManager(
+        ...     conn, default_database=DBNAME, root_database=DBNAME)
+
+        >>> foo3_B = dm_B.load(foo3_ref)
+        >>> foo3_B.name = '3'
+        >>> dm_B.tpc_finish(None)
+
+        >>> tuple(coll.find({}))
+        ({u'_id': ObjectId('4f5c114f37a08e2cac000000'), u'name': u'one'},
+         {u'_id': ObjectId('4f5c114f37a08e2cac000001'), u'name': u'two'},
+         {u'_id': ObjectId('4f5c114f37a08e2cac000002'), u'name': u'3'})
+
+    3. Transaction A modifies Foo 1 and the data is flushed:
+
+        >>> foo1_A.name = '1'
+        >>> dm.flush()
+
+        >>> tuple(coll.find({}))
+        ({u'_id': ObjectId('4f5c114f37a08e2cac000000'), u'name': u'1'},
+         {u'_id': ObjectId('4f5c114f37a08e2cac000001'), u'name': u'two'},
+         {u'_id': ObjectId('4f5c114f37a08e2cac000002'), u'name': u'3'})
+
+    4. If transcation A is later aborted, only objects modified within the
+       transaction get reset to their original state (and not all loaded ones:
+
+       >>> dm.abort(None)
+
+        >>> tuple(coll.find({}))
+        ({u'_id': ObjectId('4f5c114f37a08e2cac000000'), u'name': u'one'},
+         {u'_id': ObjectId('4f5c114f37a08e2cac000001'), u'name': u'two'},
+         {u'_id': ObjectId('4f5c114f37a08e2cac000002'), u'name': u'3'})
+    """
+
+def doctest_MongoDataManager_abort_conflict_detection():
+    r"""MongoDataManager: abort(): Conflict detections while aborting.
+
+    When a transaction is aborting, we are usually resetting the state of the
+    modified objects. What happens, however, when the document was updated
+    since the last flush?
+
+    The implemented policy now does not reset the state in this case.
+
+    First let's create an initial state:
+
+      >>> dm.conflict_handler = conflict.SimpleSerialConflictHandler(dm)
       >>> dm.reset()
-      >>> dm._registered_objects = [foo]
-      >>> dm.commit(transaction.get())
-
-    Next, let's cause a conflict byt simulating a conflicting transaction:
-
+      >>> foo_ref = dm.insert(Foo('one'))
       >>> dm.reset()
-      >>> foo2 = dm.load(ref)
-      >>> foo2.name = 'foo2'
-      >>> transaction.commit()
+      >>> coll = dm._get_collection_from_object(Foo())
+
+    1. Transaction A loads the object and modifies it:
+
+       >>> foo_A = dm.load(foo_ref)
+       >>> foo_A.name = u'1'
+       >>> coll.find_one({})
+       {u'_id': ObjectId('4e7ddf12e138237403000000'),
+        u'_py_serial': 1, u'name': u'one'}
+
+    2. Transaction B comes along and modifies the object as well and commits:
+
+       >>> dm_B = datamanager.MongoDataManager(
+       ...     conn,
+       ...     default_database=DBNAME, root_database=DBNAME,
+       ...     conflict_handler_factory=conflict.SimpleSerialConflictHandler)
+
+       >>> foo_B = dm_B.load(foo_ref)
+       >>> foo_B.name = 'Eins'
+       >>> dm_B.tpc_finish(None)
+       >>> coll.find_one({})
+       {u'_id': ObjectId('4e7ddf12e138237403000000'), u'_py_serial': 2,
+        u'name': u'Eins'}
+
+    3. If transcation A is later aborted, it does not reset the state, since
+       it changed:
+
+       >>> dm.abort(None)
+       >>> coll.find_one({})
+       {u'_id': ObjectId('4e7ddf12e138237403000000'), u'_py_serial': 2,
+        u'name': u'Eins'}
 
-      >>> dm.reset()
-      >>> dm._registered_objects = [foo]
-      >>> dm.commit(transaction.get())
-      Traceback (most recent call last):
-      ...
-      ConflictError: database conflict error
-          (oid DBRef('mongopersist.tests.test_datamanager.Foo',
-                     ObjectId('4eb3499637a08e1c5a000000'),
-                     'mongopersist_test'),
-           class Foo, start serial 1, current serial 2)
     """
 
 def doctest_MongoDataManager_tpc_begin():
@@ -598,7 +720,7 @@ def doctest_MongoDataManager_tpc_finish():
     This method finishes the two-phase commit. So let's store a simple object:
 
       >>> foo = Foo()
-      >>> dm.detect_conflicts = True
+      >>> dm.conflict_handler = conflict.SimpleSerialConflictHandler(dm)
       >>> dm._registered_objects = [foo]
       >>> dm.tpc_finish(transaction.get())
       >>> foo._p_serial
@@ -786,6 +908,59 @@ def doctest_ProcessSpecDecorator_basic():
       >>> PlacelessSetup().tearDown()
     """
 
+
+def doctest_LoggingDecorator_basic():
+    r"""class LoggingDecorator: basic
+
+    The ``LoggingDecorator`` decorator will log the name, arguments ans even
+    current stack of a function call. Let's stub the logger:
+
+      >>> orig_log_debug = datamanager.COLLECTION_LOG.debug
+      >>> def fake_debug(msg, *args):
+      ...     print msg % args
+      >>> datamanager.COLLECTION_LOG.debug = fake_debug
+
+    Let's create the decorator:
+
+      >>> coll = conn[DBNAME]['mongopersist.tests.test_datamanager.Foo']
+      >>> logging_find = datamanager.LoggingDecorator(coll, coll.find)
+      >>> list(logging_find({'life': 42}))
+      collection: mongopersist_test.mongopersist.tests.test_datamanager.Foo find,
+       TXN:('... - ',),
+       args:({'life': 42},),
+       kwargs:{},
+       tb:
+          ...
+          list(logging_find({'life': 42}))
+      <BLANKLINE>
+      []
+
+    Keyword arguments are also supported:
+
+      >>> list(logging_find(spec={'life': 42}))
+      collection: mongopersist_test.mongopersist.tests.test_datamanager.Foo find,
+       TXN:('... - ',),
+       args:(),
+       kwargs:{'spec': {'life': 42}},
+       tb:
+          ...
+          list(logging_find(spec={'life': 42}))
+      <BLANKLINE>
+      []
+
+    Tracebacks can also be turned off:
+
+      >>> logging_find.ADD_TB = False
+      >>> list(logging_find({'life': 42}))
+      collection: mongopersist_test.mongopersist.tests.test_datamanager.Foo find,
+       TXN:('... - ',),
+       args:({'life': 42},),
+       kwargs:{},
+       tb:
+        <omitted>
+      []
+    """
+
 def test_suite():
     return doctest.DocTestSuite(
         setUp=testing.setUp, tearDown=testing.tearDown,
