@@ -15,7 +15,9 @@
 import UserDict
 import persistent
 import pymongo.dbref
+import pymongo.objectid
 import zope.component
+from bson.errors import InvalidId
 from rwproperty import getproperty, setproperty
 from zope.container import contained, sample
 from zope.container.interfaces import IContainer
@@ -149,29 +151,38 @@ class MongoContainer(contained.Contained,
             filter[self._m_parent_key] = gs(self._m_get_parent_key_value())
         return filter
 
+    def _m_add_items_filter(self, filter):
+        for key, value in self._m_get_items_filter().items():
+            if key not in filter:
+                filter[key] = value
+
+    def _locate(self, obj, doc):
+        obj._v_key = doc[self._m_mapping_key]
+        obj._v_parent = self
+
     def _load_one(self, doc):
         # Create a DBRef object and then load the full state of the object.
         dbref = pymongo.dbref.DBRef(
             self._m_collection, doc['_id'],
             self._m_database or self._m_jar.default_database)
-        obj = self._m_jar._reader.get_ghost(dbref)
-        self._m_jar.setstate(obj, doc)
-        obj._v_key = doc[self._m_mapping_key]
-        obj._v_parent = self
+        # Stick the doc into the _latest_states:
+        self._m_jar._latest_states[dbref] = doc
+        obj = self._m_jar.load(dbref)
+        self._locate(obj, doc)
         return obj
 
     def __getitem__(self, key):
         filter = self._m_get_items_filter()
         filter[self._m_mapping_key] = key
-        coll = self.get_collection()
-        doc = coll.find_one(filter)
-        if doc is None:
+        obj = self.find_one(filter)
+        if obj is None:
             raise KeyError(key)
-        return self._load_one(doc)
+        return obj
 
     def _real_setitem(self, key, value):
         # This call by iteself caues the state to change _p_changed to True.
-        setattr(value, self._m_mapping_key, key)
+        if self._m_mapping_key is not None:
+            setattr(value, self._m_mapping_key, key)
         if self._m_parent_key is not None:
             setattr(value, self._m_parent_key, self._m_get_parent_key_value())
 
@@ -205,17 +216,29 @@ class MongoContainer(contained.Contained,
         # Send the uncontained event.
         contained.uncontained(value, self, key)
 
+    def __contains__(self, key):
+        return self.raw_find_one(
+            {self._m_mapping_key: key}, fields=()) is not None
+
+    def __iter__(self):
+        result = self.raw_find(
+            {self._m_mapping_key: {'$ne': None}}, fields=(self._m_mapping_key,))
+        for doc in result:
+            yield doc[self._m_mapping_key]
+
     def keys(self):
-        filter = self._m_get_items_filter()
-        filter[self._m_mapping_key] = {'$ne': None}
-        coll = self.get_collection()
-        return [doc[self._m_mapping_key]
-                for doc in coll.find(filter, fields=(self._m_mapping_key,))]
+        return list(self.__iter__())
+
+    def iteritems(self):
+        result = self.raw_find()
+        for doc in result:
+            obj = self._load_one(doc)
+            yield doc[self._m_mapping_key], obj
 
     def raw_find(self, spec=None, *args, **kwargs):
         if spec is None:
             spec  = {}
-        spec.update(self._m_get_items_filter())
+        self._m_add_items_filter(spec)
         coll = self.get_collection()
         return coll.find(spec, *args, **kwargs)
 
@@ -231,7 +254,7 @@ class MongoContainer(contained.Contained,
             spec_or_id  = {}
         if not isinstance(spec_or_id, dict):
             spec_or_id = {'_id': spec_or_id}
-        spec_or_id.update(self._m_get_items_filter())
+        self._m_add_items_filter(spec_or_id)
         coll = self.get_collection()
         return coll.find_one(spec_or_id, *args, **kwargs)
 
@@ -241,6 +264,53 @@ class MongoContainer(contained.Contained,
             return None
         return self._load_one(doc)
 
+
+class IdNamesMongoContainer(MongoContainer):
+    """A container that uses the Mongo ObjectId as the name/key."""
+    _m_mapping_key = None
+
+
+    @property
+    def _m_remove_documents(self):
+        # Objects must be removed, since removing the _id of a document is not
+        # allowed.
+        return True
+
+    def _locate(self, obj, doc):
+        obj._v_key = unicode(doc['_id'])
+        obj._v_parent = self
+
+    def __getitem__(self, key):
+        try:
+            id = pymongo.objectid.ObjectId(key)
+        except InvalidId:
+            raise KeyError(key)
+        filter = self._m_get_items_filter()
+        filter['_id'] = id
+        obj = self.find_one(filter)
+        if obj is None:
+            raise KeyError(key)
+        return obj
+
+    def __contains__(self, key):
+        try:
+            id = pymongo.objectid.ObjectId(key)
+        except InvalidId:
+            return False
+        return self.raw_find_one({'_id': id}, fields=()) is not None
+
+    def __iter__(self):
+        result = self.raw_find(fields=None)
+        for doc in result:
+            yield unicode(doc['_id'])
+
+    def iteritems(self):
+        result = self.raw_find()
+        for doc in result:
+            obj = self._load_one(doc)
+            yield unicode(doc['_id']), obj
+
+
 class AllItemsMongoContainer(MongoContainer):
     _m_parent_key = None
 
